"""
Testing for AGE-ABM Backend (backend, visual and utils)
@Author Max Hall
@Co-Authors Meghan Ireland and Matthew Fleischman

"""

import subprocess as sp
import backend.backend as bk
import backend.backendutils as bk_utils


class TestInitialisation:
    ''' Ensures that the class is of the correct type and initialises correctly '''

    def test__init__(self):
        ''' Ensures that the class is of the correct type and initialises correctly '''
        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        assert str(type(writingfile)) == "<class 'backend.backendutils.WriteModelFile'>"
        assert writingfile.file_name == "initialisingtesting"
        assert writingfile.imports_section == ""
        assert writingfile.components_section == ""
        assert writingfile.agents_section == ""
        assert writingfile.systems_section == ""
        assert writingfile.data_collection_section == ""
        assert writingfile.model_section == ""
        assert writingfile.main_file_section == ""


class TestCreatingCodeSections:
    ''' Ensures code to be written to files is generated properly '''
    def test_create_imports_section(self):
        ''' Tests that the adding of imports section works '''
        model = bk.ModelEnv("codesectiontesting")
        model.add_import("pytest", "pytests", "from", ["test1", "test2"])
        model.add_import("pytorch", "torch", "*")
        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        response = writingfile.create_imports_section(model)
        assert response == 0


    def test_create_components_section(self):
        ''' Tests that the adding of components section works '''
        model = bk.ModelEnv("codesectiontesting")
        model.add_component("Comp1", "This is comp")
        model.add_attribute("Comp1", "attr1", "bool", "agent_lvl")
        model.add_attribute("Comp1", "attr2", "bool", "model_lvl")
        model.add_attribute("Comp1", "attr2", "bool", "comp_lvl", True)

        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        response = writingfile.create_components_section(model)
        assert response == 0


    def test_create_agents_section(self):
        ''' Tests that the adding of agents section works '''
        model = bk.ModelEnv("codesectiontesting")
        model.add_component("Comp1", "This is comp")
        model.add_component("Comp2", "This is comp", True)
        model.add_attribute("Comp1", "attr1", "bool", "agent_lvl")
        model.add_attribute("Comp1", "attr2", "bool", "model_lvl")
        model.add_attribute("Comp1", "attr2", "bool", "comp_lvl", True)
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent2", "This is agent", ["Comp1"])
        model.add_agent("Agent3", "This is agent", ["Comp1"])
        model.add_agent_lvl_default_value("Agent1", "Comp1", "attr1", "return 1")
        model.add_agent_lvl_default_value("Agent2", "Comp1", "attr1", "return 1")
        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        # Agent level attribute initial value not added
        response = writingfile.create_agents_section(model)
        assert response == 1

        model.remove_agent("Agent3")
        response = writingfile.create_agents_section(model)
        assert response == 0


    def test_create_systems_section(self):
        ''' Tests that the adding of systems section works '''
        model = bk.ModelEnv("codesectiontesting")
        model.add_system("System1", "This is system", start = 3, end = 100, frequency = 2)
        model.add_system("System2", "This is system")
        model.add_generator("System1", "generator1", "This is generator", nickname = "gen1")
        model.add_generator("System1", "generator2", "This is generator", nickname = "gen2", return_value = "return 3 if cool else 2")
        model.add_gen_attribute("System1", "Generator1", "Gen Att 1", "int")
        model.add_gen_attribute("System1", "Generator1", "Gen Att 2", "int")
        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        
        response = writingfile.create_systems_section(model)
        assert response == 0


    def test_create_data_collection_section(self):
        ''' Tests that the adding of data_collection section works '''
        model = bk.ModelEnv("validationtesting")
        model.add_component("Comp1", "This is component for add testing")
        model.add_component("Comp2", "This is component for add testing")
        model.add_component("Comp3", "This is component for add testing")
        model.add_component("Comp4", "This is component for add testing")
        model.add_component("Comp5", "This is component for add testing")
        model.add_component("Comp6", "This is component for add testing")
        model.add_attribute("Comp1", "attr1", "int", "comp_lvl", 1)
        model.add_attribute("Comp1", "attr2", "int", "comp_lvl", 1)
        model.add_attribute("Comp1", "attr3", "bool", "comp_lvl", True)
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2", "Comp3", "Comp4", "Comp5", "Comp6"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2", "Comp3", "Comp6"])
        model.add_agent("Agent3", "This is agent", ["Comp3", "Comp4", "Comp5", "Comp6"])
        model.add_agent("Agent4", "This is agent", ["Comp1", "Comp2", "Comp6"])
        model.add_agent("Agent5", "This is agent", ["Comp1", "Comp2", "Comp6"])
        model.add_agent_of_interest("Agent1", ["Comp1", "Comp2", "Comp3", "Comp4", "Comp5"], count = True, individual = True, mean_val = True, median_val = True, total_val = True, max_val = True, min_val = True, variance = True, std_dev = True)
        model.add_agent_of_interest("Agent2", [], count = True)
        model.add_agent_of_interest("Agent3", ["Comp3", "Comp4", "Comp5"], count = True)
        model.add_agent_of_interest("Agent4", ["Comp1", "Comp2", "Comp6"], position = True)
        model.add_agent_of_interest("Agent5", [], position = True)
        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        response = writingfile.create_data_collection_section(model)
        assert response == 0


    def test_create_model_section(self):
        ''' Tests that the adding of model section works '''
        model = bk.ModelEnv("validationtesting")
        model.add_component("Comp1", "This is component for add testing", True)
        model.add_component("Comp2", "This is component for add testing")
        model.add_component("Comp3", "This is component for add testing")
        model.add_component("Comp4", "This is component for add testing")
        model.add_component("Comp5", "This is component for add testing")
        model.add_component("Comp6", "This is component for add testing")
        model.add_attribute("Comp1", "attr1", "int", "model_lvl")
        model.add_attribute("Comp1", "attr2", "int", "model_lvl")
        model.add_attribute("Comp1", "attr3", "bool", "model_lvl")
        model.add_attribute("Comp2", "attr1", "int", "model_lvl")
        model.add_attribute("Comp2", "attr2", "int", "model_lvl")
        model.add_attribute("Comp2", "attr3", "bool", "model_lvl")
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2", "Comp3", "Comp4", "Comp5", "Comp6"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2", "Comp3", "Comp6"])
        model.add_system("System1", "This is system", start = 3, end = 100, frequency = 2)
        model.add_system("System2", "This is system")
        model.add_generator("System1", "generator1", "This is generator", nickname = "gen1", return_value = "return 5 if cool else 2")
        model.add_generator("System1", "generator2", "This is generator", nickname = "gen2", return_value = "return 3 if cool else 2")
        model.add_gen_attribute("System1", "Generator1", "Gen Att 1", "int")
        model.add_gen_attribute("System1", "Generator1", "Gen Att 2", "int")
        model.add_gen_attribute("System1", "Generator2", "Gen Att 1", "int")

        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        response = writingfile.create_model_section(model)
        assert response == 0

        # Gridworld cases
        model.get_model_details().edit_gridworld(True)
        
        # Agents without initial co-ordinates
        response = writingfile.create_model_section(model)
        assert response == 1

        model.edit_agent("Agent1", init_coordinates = ["1", "2"])
        model.edit_agent("Agent2", init_coordinates = ["2", "3"])
        response = writingfile.create_model_section(model)
        assert response == 0


    def test_create_main_file_section(self):
        ''' Tests that the adding of main file section works '''
        model = bk.ModelEnv("codesectiontesting")
        model.add_component("Comp1", "This is component for add testing", True)
        model.add_component("Comp2", "This is component for add testing")
        model.add_component("Comp3", "This is component for add testing")
        model.add_component("Comp4", "This is component for add testing")
        model.add_component("Comp5", "This is component for add testing")
        model.add_component("Comp6", "This is component for add testing")
        model.add_attribute("Comp1", "attr1", "int", "model_lvl")
        model.add_attribute("Comp1", "attr2", "int", "model_lvl")
        model.add_attribute("Comp1", "attr3", "bool", "model_lvl")
        model.add_attribute("Comp2", "attr1", "int", "model_lvl")
        model.add_attribute("Comp2", "attr2", "int", "model_lvl")
        model.add_attribute("Comp2", "attr3", "bool", "model_lvl")
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2", "Comp3", "Comp4", "Comp5", "Comp6"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2", "Comp3", "Comp6"])
        model.add_system("System1", "This is system", start = 3, end = 100, frequency = 2)
        model.add_system("System2", "This is system")
        model.add_generator("System1", "generator1", "This is generator", nickname = "gen1", return_value = "return 5 if cool else 2")
        model.add_generator("System1", "generator2", "This is generator", nickname = "gen2", return_value = "return 3 if cool else 2")
        model.add_gen_attribute("System1", "Generator1", "Gen Att 1", "int")
        model.add_gen_attribute("System1", "Generator1", "Gen Att 2", "int")
        model.add_gen_attribute("System1", "Generator2", "Gen Att 1", "int")

        writingfile = bk_utils.WriteModelFile("initialisingtesting")
        response = writingfile.create_main_file_section(model, visualise_model = False)
        assert response == 0
 
        response = writingfile.create_main_file_section(model, visualise_model = True)
        assert response == 0


class TestWritingToFiles:
    ''' Ensures code to be written to files is generated properly '''

    def test_create_model_file(self):
        ''' Tests that creating the model file works '''
        model = bk.ModelEnv("codewritingtesting")
        writingfile = bk_utils.WriteModelFile("writingtesting")
        writingfile.create_imports_section(model)
        writingfile.create_components_section(model)
        writingfile.create_agents_section(model)
        writingfile.create_systems_section(model)
        writingfile.create_data_collection_section(model)
        writingfile.create_model_section(model)

        response = writingfile.create_model_file()
        assert response == 0

    
    def test_create_main_file(self):
        ''' Tests that creating the main file works '''
        model = bk.ModelEnv("codewritingtesting")
        writingfile = bk_utils.WriteModelFile("writingtesting")
        writingfile.create_imports_section(model)
        writingfile.create_components_section(model)
        writingfile.create_agents_section(model)
        writingfile.create_systems_section(model)
        writingfile.create_data_collection_section(model)
        writingfile.create_model_section(model)
        
        writingfile.create_main_file_section(model, visualise_model=False)
        response = writingfile.create_main_file()
        assert response == 0

        sp.run(["rm", "-rf", "runtimeFiles"])

        writingfile.create_main_file_section(model, visualise_model=True)
        response = writingfile.create_main_file()
        assert response == 0
