"""
Testing for AGE-ABM Backend (backend, visual and utils)
@Author Max Hall
@Co-Authors Meghan Ireland and Matthew Fleischman

"""

''' I make use of jsonpickle for testing get as it breaks the class into a string to allow for comparison '''
from jsonpickle import dumps
import visuals.visual as vis
import backend.backend as bk
import visuals.visual_classes as vis_classes


class TestInitialisation:
    ''' Ensures that the class is of the correct type and initialises correctly '''
    def test__init__(self):
        ''' Ensures that the class is of the correct type and initialises correctly '''
        model = bk.ModelEnv("tesingvis")
        vismodel = vis.VisualEnv(model)
        assert str(type(vismodel)) == "<class 'visuals.visual.VisualEnv'>"

        assert dumps(vismodel.model) == dumps(bk.ModelEnv("tesingvis"))
        assert vismodel.visuals == {}


class TestVisualiseAgentMethods:
    ''' Tests for Visualise Agent methods '''
    def test_add_agent_visualisation(self):
        ''' Tests that the adding of agent of visualisation works '''
        model = bk.ModelEnv("tesingvis")
        vismodel = vis.VisualEnv(model)
        model.add_component("Comp1", "This is component for add testing")
        model.add_component("Comp2", "This is component for add testing")
        model.add_attribute("Comp1", "Attr1", "int", "comp_lvl", 1)
        model.add_attribute("Comp1", "Attr2", "bool", "comp_lvl", True)
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent3", "This is agent", ["Comp1"])
        model.add_agent_of_interest("Agent1", ["Comp1", "Comp2"], count = True, mean_val=True, individual = True, max_val = True)
        model.add_agent_of_interest("Agent2", ["Comp1", "Comp2"], count = True, mean_val=True, position = True, max_val = True)
        model.add_agent_of_interest("Agent3", ["Comp1"], count = True, mean_val=True, individual = True, max_val = True)

        # Valid input
        response = vismodel.add_agent_visualisation("Plot1", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr1")
        assert response == 0
        response = vismodel.add_agent_visualisation("Plot2", ["Agent1"], "mean_val", "plot", "Comp1", "Attr1")
        assert response == 0

        # Plot name already in use
        response = vismodel.add_agent_visualisation("Plot1", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr1")
        assert response == 1

        # Comp name doesn't exist
        response = vismodel.add_agent_visualisation("Plot3", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp3", "Attr1")
        assert response == 1
        
        # Attr name doesn't exist
        response = vismodel.add_agent_visualisation("Plot3", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr3")
        assert response == 1

        # Invalid plot name
        response = vismodel.add_agent_visualisation("Plot@1", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr1")
        assert response == 2

        # No agents provided
        response = vismodel.add_agent_visualisation("Plot3", [], "mean_val", "plot", "Comp1", "Attr1")
        assert response == 3

        # Invalid data type
        response = vismodel.add_agent_visualisation("Plot3", ["Agent1", "Agent2", "Agent3"], "cows", "plot", "Comp1", "Attr1")
        assert response == 4

        # Data type not shared between given agents
        response = vismodel.add_agent_visualisation("Plot3", ["Agent1", "Agent2", "Agent3"], "individual", "plot", "Comp1", "Attr1")
        assert response == 5

        # Plot type is wrong type
        response = vismodel.add_agent_visualisation("Plot3", ["Agent1", "Agent2", "Agent3"], "mean_val", "plt", "Comp1", "Attr1")
        assert response == 6


    def test_get_agent_visualisation(self):
        ''' Tests that the getting of agent of visualisation works '''
        model = bk.ModelEnv("tesingvis")
        vismodel = vis.VisualEnv(model)
        model.add_component("Comp1", "This is component for add testing")
        model.add_component("Comp2", "This is component for add testing")
        model.add_attribute("Comp1", "Attr1", "int", "comp_lvl", 1)
        model.add_attribute("Comp1", "Attr2", "bool", "comp_lvl", True)
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent3", "This is agent", ["Comp1"])
        model.add_agent_of_interest("Agent1", ["Comp1", "Comp2"], count = True, mean_val=True, individual = True, max_val = True)
        model.add_agent_of_interest("Agent2", ["Comp1", "Comp2"], count = True, mean_val=True, position = True, max_val = True)
        model.add_agent_of_interest("Agent3", ["Comp1"], count = True, mean_val=True, individual = True, max_val = True)
        vismodel.add_agent_visualisation("Plot1", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr1")

        response = vismodel.get_agent_visualisation("Plot1")
        assert dumps(response) == dumps(vis_classes.VisualiseAgent("Agent1", ["Agent2", "Agent3"], "mean_val", "plot", "Comp1", "attr1"))

        # Name Errors
        response = vismodel.get_agent_visualisation("Plot2")
        assert response == None
        response = vismodel.get_agent_visualisation("Pl#$ot1")
        assert response == None


    def test_remove_agent_visualisation(self):
        ''' Tests that the removing of agent of visualisation works '''
        
        model = bk.ModelEnv("tesingvis")
        vismodel = vis.VisualEnv(model)
        model.add_component("Comp1", "This is component for add testing")
        model.add_component("Comp2", "This is component for add testing")
        model.add_attribute("Comp1", "Attr1", "int", "comp_lvl", 1)
        model.add_attribute("Comp1", "Attr2", "bool", "comp_lvl", True)
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent3", "This is agent", ["Comp1"])
        model.add_agent_of_interest("Agent1", ["Comp1", "Comp2"], count = True, mean_val=True, individual = True, max_val = True)
        model.add_agent_of_interest("Agent2", ["Comp1", "Comp2"], count = True, mean_val=True, position = True, max_val = True)
        model.add_agent_of_interest("Agent3", ["Comp1"], count = True, mean_val=True, individual = True, max_val = True)
        vismodel.add_agent_visualisation("Plot1", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr1")

        # Invalid name
        response = vismodel.remove_agent_visualisation("Plot@#1")
        assert response == 1

        # Valid name
        response = vismodel.remove_agent_visualisation("Plot1")
        assert response == 0

        # Invalid names
        response = vismodel.remove_agent_visualisation("Plot1")
        assert response == 1
        response = vismodel.remove_agent_visualisation("Plot@#1")
        assert response == 1


    def test_get_list_visuals(self):
        ''' Tests that the getting list of visuals works '''
        
        model = bk.ModelEnv("tesingvis")
        vismodel = vis.VisualEnv(model)
        model.add_component("Comp1", "This is component for add testing")
        model.add_component("Comp2", "This is component for add testing")
        model.add_attribute("Comp1", "Attr1", "int", "comp_lvl", 1)
        model.add_attribute("Comp1", "Attr2", "bool", "comp_lvl", True)
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent3", "This is agent", ["Comp1"])
        model.add_agent_of_interest("Agent1", ["Comp1", "Comp2"], count = True, mean_val=True, individual = True, max_val = True)
        model.add_agent_of_interest("Agent2", ["Comp1", "Comp2"], count = True, mean_val=True, position = True, max_val = True)
        model.add_agent_of_interest("Agent3", ["Comp1"], count = True, mean_val=True, individual = True, max_val = True)
        vismodel.add_agent_visualisation("Plot1", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr1")
        vismodel.add_agent_visualisation("Plot2", ["Agent1", "Agent2"], "mean_val", "plot", "Comp1", "Attr1")
        vismodel.add_agent_visualisation("Plot3", ["Agent1", "Agent2", "Agent3"], "count", "plot")

        response = vismodel.get_list_visuals()
        assert set(response) == set(["Plot1", "Plot2", "Plot3"])
        vismodel.visuals = {}
        response = vismodel.get_list_visuals()
        assert len(response) == 0


class TestFilingMethods:
    ''' Tests for Filing methods '''
    def test_visualise_model(self):
        ''' Tests that creating the visualisation file works '''
        model = bk.ModelEnv("tesingvis")
        vismodel = vis.VisualEnv(model)
        model.add_component("Comp1", "This is component for add testing")
        model.add_component("Comp2", "This is component for add testing")
        model.add_attribute("Comp1", "Attr1", "int", "comp_lvl", 1)
        model.add_attribute("Comp1", "Attr2", "bool", "comp_lvl", True)
        model.add_agent("Agent1", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent2", "This is agent", ["Comp1", "Comp2"])
        model.add_agent("Agent3", "This is agent", ["Comp1"])
        model.add_agent_of_interest("Agent1", ["Comp1", "Comp2"], count = True, mean_val=True, individual = True, max_val = True)
        model.add_agent_of_interest("Agent2", ["Comp1", "Comp2"], count = True, mean_val=True, position = True, max_val = True)
        model.add_agent_of_interest("Agent3", ["Comp1"], count = True, mean_val=True, individual = True, max_val = True)
        vismodel.add_agent_visualisation("Plot1", ["Agent1", "Agent2", "Agent3"], "mean_val", "plot", "Comp1", "Attr1")
        vismodel.add_agent_visualisation("Plot2", ["Agent1", "Agent2"], "mean_val", "plot", "Comp1", "Attr1")
        vismodel.add_agent_visualisation("Plot3", ["Agent1", "Agent2", "Agent3"], "count", "plot")

        response = vismodel.visualise_model("Plot1")
        assert response == 0
        response = vismodel.visualise_model("Plot2.py")
        assert response == 0
        response = vismodel.visualise_model("Plot4.py")
        assert response == 1
        response = vismodel.visualise_model("Plo@4.py")
        assert response == 1